@@ -1,5 +1,5 @@
 /*
-* Copyright (c) 2013-2023, The PurpleI2P Project
+* Copyright (c) 2013-2024, The PurpleI2P Project
 *
 * This file is part of Purple i2pd project and licensed under BSD3
 *
@@ -106,7 +106,7 @@ namespace data
 	{
 		i2p::util::SetThreadName("NetDB");
 
-		uint64_t lastManage = 0, lastExploratory = 0, lastManageRequest = 0, lastDestinationCleanup = 0;
+		uint64_t lastManage = 0, lastExploratory = 0, lastManageRequest = 0;
 		uint64_t lastProfilesCleanup = i2p::util::GetSecondsSinceEpoch ();
 		int16_t profilesCleanupVariance = 0;
 
@@ -149,7 +149,7 @@ namespace data
 				if (!i2p::transport::transports.IsOnline ()) continue; // don't manage netdb when offline
 
 				uint64_t ts = i2p::util::GetSecondsSinceEpoch ();
-				if (ts - lastManageRequest >= 15 || ts + 15 < lastManageRequest) // manage requests every 15 seconds
+				if (ts - lastManageRequest >= MANAGE_REQUESTS_INTERVAL || ts + MANAGE_REQUESTS_INTERVAL < lastManageRequest) // manage requests every 15 seconds
 				{
 					m_Requests.ManageRequests ();
 					lastManageRequest = ts;
@@ -165,13 +165,6 @@ namespace data
 					lastManage = ts;
 				}
 
-				if (ts - lastDestinationCleanup >= i2p::garlic::INCOMING_TAGS_EXPIRATION_TIMEOUT ||
-				    ts + i2p::garlic::INCOMING_TAGS_EXPIRATION_TIMEOUT < lastDestinationCleanup)
-				{
-					i2p::context.CleanupDestination ();
-					lastDestinationCleanup = ts;
-				}
-
 				if (ts - lastProfilesCleanup >= (uint64_t)(i2p::data::PEER_PROFILE_AUTOCLEAN_TIMEOUT + profilesCleanupVariance) ||
 				    ts + i2p::data::PEER_PROFILE_AUTOCLEAN_TIMEOUT < lastProfilesCleanup)
 				{
@@ -226,7 +219,8 @@ namespace data
 	bool NetDb::AddRouterInfo (const IdentHash& ident, const uint8_t * buf, int len)
 	{
 		bool updated;
-		AddRouterInfo (ident, buf, len, updated);
+		if (!AddRouterInfo (ident, buf, len, updated)) 
+			updated = false;
 		return updated;
 	}
 
@@ -240,31 +234,34 @@ namespace data
 			{
 				bool wasFloodfill = r->IsFloodfill ();
 				{
-					std::unique_lock<std::mutex> l(m_RouterInfosMutex);
+					std::lock_guard<std::mutex> l(m_RouterInfosMutex);
 					if (!r->Update (buf, len))
 					{
 						updated = false;
 						m_Requests.RequestComplete (ident, r);
 						return r;
 					}
-					if (r->IsUnreachable ())
+					if (r->IsUnreachable () ||
+					    i2p::util::GetMillisecondsSinceEpoch () + NETDB_EXPIRATION_TIMEOUT_THRESHOLD*1000LL < r->GetTimestamp ())
 					{
-						// delete router as invalid after update
+						// delete router as invalid or from future after update
 						m_RouterInfos.erase (ident);
 						if (wasFloodfill)
 						{
-							std::unique_lock<std::mutex> l(m_FloodfillsMutex);
+							std::lock_guard<std::mutex> l(m_FloodfillsMutex);
 							m_Floodfills.Remove (r->GetIdentHash ());
 						}
 						m_Requests.RequestComplete (ident, nullptr);
 						return nullptr;
 					}
 				}
-				LogPrint (eLogInfo, "NetDb: RouterInfo updated: ", ident.ToBase64());
+				if (CheckLogLevel (eLogInfo))
+					LogPrint (eLogInfo, "NetDb: RouterInfo updated: ", ident.ToBase64());
 				if (wasFloodfill != r->IsFloodfill ()) // if floodfill status updated
 				{
-					LogPrint (eLogDebug, "NetDb: RouterInfo floodfill status updated: ", ident.ToBase64());
-					std::unique_lock<std::mutex> l(m_FloodfillsMutex);
+					if (CheckLogLevel (eLogDebug))
+						LogPrint (eLogDebug, "NetDb: RouterInfo floodfill status updated: ", ident.ToBase64());
+					std::lock_guard<std::mutex> l(m_FloodfillsMutex);
 					if (wasFloodfill)
 						m_Floodfills.Remove (r->GetIdentHash ());
 					else if (r->IsEligibleFloodfill ())
@@ -278,7 +275,8 @@ namespace data
 			}
 			else
 			{
-				LogPrint (eLogDebug, "NetDb: RouterInfo is older: ", ident.ToBase64());
+				if (CheckLogLevel (eLogDebug))
+					LogPrint (eLogDebug, "NetDb: RouterInfo is older: ", ident.ToBase64());
 				updated = false;
 			}
 		}
@@ -290,18 +288,19 @@ namespace data
 			{
 				bool inserted = false;
 				{
-					std::unique_lock<std::mutex> l(m_RouterInfosMutex);
+					std::lock_guard<std::mutex> l(m_RouterInfosMutex);
 					inserted = m_RouterInfos.insert ({r->GetIdentHash (), r}).second;
 				}
 				if (inserted)
 				{
-					LogPrint (eLogInfo, "NetDb: RouterInfo added: ", ident.ToBase64());
+					if (CheckLogLevel (eLogInfo))
+						LogPrint (eLogInfo, "NetDb: RouterInfo added: ", ident.ToBase64());
 					if (r->IsFloodfill () && r->IsEligibleFloodfill ())
 					{
 						if (m_Floodfills.GetSize () < NETDB_NUM_FLOODFILLS_THRESHOLD ||
 						 r->GetProfile ()->IsReal ()) // don't insert floodfill until it's known real if we have enough
 						{
-							std::unique_lock<std::mutex> l(m_FloodfillsMutex);
+							std::lock_guard<std::mutex> l(m_FloodfillsMutex);
 							m_Floodfills.Insert (r);
 						}
 						else
@@ -324,7 +323,7 @@ namespace data
 
 	bool NetDb::AddLeaseSet (const IdentHash& ident, const uint8_t * buf, int len)
 	{
-		std::unique_lock<std::mutex> lock(m_LeaseSetsMutex);
+		std::lock_guard<std::mutex> lock(m_LeaseSetsMutex);
 		bool updated = false;
 		auto it = m_LeaseSets.find(ident);
 		if (it != m_LeaseSets.end () && it->second->GetStoreType () == i2p::data::NETDB_STORE_TYPE_LEASESET)
@@ -336,10 +335,11 @@ namespace data
 				if(it->second->GetExpirationTime() < expires)
 				{
 					it->second->Update (buf, len, false); // signature is verified already
-					LogPrint (eLogInfo, "NetDb: LeaseSet updated: ", ident.ToBase32());
+					if (CheckLogLevel (eLogInfo))
+						LogPrint (eLogInfo, "NetDb: LeaseSet updated: ", ident.ToBase32());
 					updated = true;
 				}
-				else
+				else if (CheckLogLevel (eLogDebug))
 					LogPrint(eLogDebug, "NetDb: LeaseSet is older: ", ident.ToBase32());
 			}
 			else
@@ -350,7 +350,8 @@ namespace data
 			auto leaseSet = std::make_shared<LeaseSet> (buf, len, false); // we don't need leases in netdb
 			if (leaseSet->IsValid ())
 			{
-				LogPrint (eLogInfo, "NetDb: LeaseSet added: ", ident.ToBase32());
+				if (CheckLogLevel (eLogInfo))
+					LogPrint (eLogInfo, "NetDb: LeaseSet added: ", ident.ToBase32());
 				m_LeaseSets[ident] = leaseSet;
 				updated = true;
 			}
@@ -365,7 +366,7 @@ namespace data
 		auto leaseSet = std::make_shared<LeaseSet2> (storeType, buf, len, false); // we don't need leases in netdb
 		if (leaseSet->IsValid ())
 		{
-			std::unique_lock<std::mutex> lock(m_LeaseSetsMutex);
+			std::lock_guard<std::mutex> lock(m_LeaseSetsMutex);
 			auto it = m_LeaseSets.find(ident);
 			if (it == m_LeaseSets.end () || it->second->GetStoreType () != storeType ||
 				leaseSet->GetPublishedTimestamp () > it->second->GetPublishedTimestamp ())
@@ -374,7 +375,8 @@ namespace data
 				     i2p::util::GetSecondsSinceEpoch () + NETDB_EXPIRATION_TIMEOUT_THRESHOLD > leaseSet->GetPublishedTimestamp ())
 				{
 					// TODO: implement actual update
-					LogPrint (eLogInfo, "NetDb: LeaseSet2 updated: ", ident.ToBase32());
+					if (CheckLogLevel (eLogInfo))
+						LogPrint (eLogInfo, "NetDb: LeaseSet2 updated: ", ident.ToBase32());
 					m_LeaseSets[ident] = leaseSet;
 					return true;
 				}
@@ -392,7 +394,7 @@ namespace data
 
 	std::shared_ptr<RouterInfo> NetDb::FindRouter (const IdentHash& ident) const
 	{
-		std::unique_lock<std::mutex> l(m_RouterInfosMutex);
+		std::lock_guard<std::mutex> l(m_RouterInfosMutex);
 		auto it = m_RouterInfos.find (ident);
 		if (it != m_RouterInfos.end ())
 			return it->second;
@@ -402,7 +404,7 @@ namespace data
 
 	std::shared_ptr<LeaseSet> NetDb::FindLeaseSet (const IdentHash& destination) const
 	{
-		std::unique_lock<std::mutex> lock(m_LeaseSetsMutex);
+		std::lock_guard<std::mutex> lock(m_LeaseSetsMutex);
 		auto it = m_LeaseSets.find (destination);
 		if (it != m_LeaseSets.end ())
 			return it->second;
@@ -436,7 +438,7 @@ namespace data
 		auto r = FindRouter (ident);
 		if (r)
 		{
-			std::unique_lock<std::mutex> l(m_RouterInfosMutex);
+			std::lock_guard<std::mutex> l(m_RouterInfosMutex);
 			r->ExcludeReachableTransports (transports);
 		}
 	}
@@ -526,7 +528,7 @@ namespace data
 
 	void NetDb::VisitLeaseSets(LeaseSetVisitor v)
 	{
-		std::unique_lock<std::mutex> lock(m_LeaseSetsMutex);
+		std::lock_guard<std::mutex> lock(m_LeaseSetsMutex);
 		for ( auto & entry : m_LeaseSets)
 			v(entry.first, entry.second);
 	}
@@ -542,7 +544,7 @@ namespace data
 
 	void NetDb::VisitRouterInfos(RouterInfoVisitor v)
 	{
-		std::unique_lock<std::mutex> lock(m_RouterInfosMutex);
+		std::lock_guard<std::mutex> lock(m_RouterInfosMutex);
 		for ( const auto & item : m_RouterInfos )
 			v(item.second);
 	}
@@ -554,7 +556,7 @@ namespace data
 		size_t iters = max_iters_per_cyle;
 		while(n > 0)
 		{
-			std::unique_lock<std::mutex> lock(m_RouterInfosMutex);
+			std::lock_guard<std::mutex> lock(m_RouterInfosMutex);
 			uint32_t idx = rand () % m_RouterInfos.size ();
 			uint32_t i = 0;
 			for (const auto & it : m_RouterInfos) {
@@ -637,6 +639,7 @@ namespace data
 					// we have something to save
 					it.second->SaveToFile (m_Storage.Path(ident));
 					it.second->SetUnreachable (false);
+					std::lock_guard<std::mutex> l(m_RouterInfosMutex); // possible collision between DeleteBuffer and Update
 					it.second->DeleteBuffer ();
 				}
 				it.second->SetUpdated (false);
@@ -669,10 +672,11 @@ namespace data
 					    (CreateRoutingKey (it.second->GetIdentHash ()) ^ i2p::context.GetIdentHash ()).metric[0] >= 0x02)) // different first 7 bits 
 							it.second->SetUnreachable (true);
 				}	
-				if (it.second->IsUnreachable () && i2p::transport::transports.IsConnected (it.second->GetIdentHash ()))
-					it.second->SetUnreachable (false); // don't expire connected router
 			}
-
+			// make router reachable back if connected now
+			if (it.second->IsUnreachable () && i2p::transport::transports.IsConnected (it.second->GetIdentHash ()))
+				it.second->SetUnreachable (false);
+			
 			if (it.second->IsUnreachable ())
 			{
 				if (it.second->IsFloodfill ()) deletedFloodfillsCount++;
@@ -695,7 +699,7 @@ namespace data
 			LogPrint (eLogInfo, "NetDb: Deleting ", deletedCount, " unreachable routers");
 			// clean up RouterInfos table
 			{
-				std::unique_lock<std::mutex> l(m_RouterInfosMutex);
+				std::lock_guard<std::mutex> l(m_RouterInfosMutex);
 				for (auto it = m_RouterInfos.begin (); it != m_RouterInfos.end ();)
 				{
 					if (!it->second || it->second->IsUnreachable ())
@@ -709,7 +713,7 @@ namespace data
 			}
 			// clean up expired floodfills or not floodfills anymore
 			{
-				std::unique_lock<std::mutex> l(m_FloodfillsMutex);
+				std::lock_guard<std::mutex> l(m_FloodfillsMutex);
 				m_Floodfills.Cleanup ([](const std::shared_ptr<RouterInfo>& r)->bool
 					{
 						return r && r->IsFloodfill () && !r->IsUnreachable ();
@@ -720,7 +724,8 @@ namespace data
 
 	void NetDb::RequestDestination (const IdentHash& destination, RequestedDestination::RequestComplete requestComplete, bool direct)
 	{
-		auto dest = m_Requests.CreateRequest (destination, false, requestComplete); // non-exploratory
+		if (direct && i2p::transport::transports.RoutesRestricted ()) direct = false; // always use tunnels for restricted routes
+		auto dest = m_Requests.CreateRequest (destination, false, direct, requestComplete); // non-exploratory
 		if (!dest)
 		{
 			LogPrint (eLogWarning, "NetDb: Destination ", destination.ToBase64(), " is requested already");
@@ -734,7 +739,11 @@ namespace data
 				!i2p::transport::transports.IsConnected (floodfill->GetIdentHash ()))
 				direct = false; // floodfill can't be reached directly
 			if (direct)
-				transports.SendMessage (floodfill->GetIdentHash (), dest->CreateRequestMessage (floodfill->GetIdentHash ()));
+			{
+				auto msg = dest->CreateRequestMessage (floodfill->GetIdentHash ());
+				msg->onDrop = [this, dest]() { this->m_Requests.SendNextRequest (dest); }; 
+				transports.SendMessage (floodfill->GetIdentHash (), msg);
+			}	
 			else
 			{
 				auto pool = i2p::tunnel::tunnels.GetExploratoryPool ();
@@ -743,6 +752,7 @@ namespace data
 				if (outbound &&	inbound)
 				{
 					auto msg = dest->CreateRequestMessage (floodfill, inbound);
+					msg->onDrop = [this, dest]() { this->m_Requests.SendNextRequest (dest); }; 
 					outbound->SendTunnelDataMsgTo (floodfill->GetIdentHash (), 0,
 						i2p::garlic::WrapECIESX25519MessageForRouter (msg, floodfill->GetIdentity ()->GetEncryptionPublicKey ()));
 				}
@@ -760,10 +770,10 @@ namespace data
 		}
 	}
 
-	void NetDb::RequestDestinationFrom (const IdentHash& destination, const IdentHash & from, bool exploritory, RequestedDestination::RequestComplete requestComplete)
+	void NetDb::RequestDestinationFrom (const IdentHash& destination, const IdentHash & from, bool exploratory, RequestedDestination::RequestComplete requestComplete)
 	{
 
-		auto dest = m_Requests.CreateRequest (destination, exploritory, requestComplete); // non-exploratory
+		auto dest = m_Requests.CreateRequest (destination, exploratory, true, requestComplete); // non-exploratory
 		if (!dest)
 		{
 			LogPrint (eLogWarning, "NetDb: Destination ", destination.ToBase64(), " is requested already");
@@ -814,17 +824,31 @@ namespace data
 			offset += 4;
 			if (replyToken != 0xFFFFFFFFU) // if not caught on OBEP or IBGW
 			{
+				IdentHash replyIdent(buf + offset);
 				auto deliveryStatus = CreateDeliveryStatusMsg (replyToken);
 				if (!tunnelID) // send response directly
-					transports.SendMessage (buf + offset, deliveryStatus);
+					transports.SendMessage (replyIdent, deliveryStatus);
 				else
 				{
-					auto pool = i2p::tunnel::tunnels.GetExploratoryPool ();
-					auto outbound = pool ? pool->GetNextOutboundTunnel () : nullptr;
-					if (outbound)
-						outbound->SendTunnelDataMsgTo (buf + offset, tunnelID, deliveryStatus);
+					bool direct = true;
+					if (!i2p::transport::transports.IsConnected (replyIdent))
+					{
+						auto r = FindRouter (replyIdent);
+						if (r && !r->IsReachableFrom (i2p::context.GetRouterInfo ()))
+							direct = false;
+					}	
+					if (direct) // send response directly to IBGW
+						transports.SendMessage (replyIdent, i2p::CreateTunnelGatewayMsg (tunnelID, deliveryStatus));
 					else
-						LogPrint (eLogWarning, "NetDb: No outbound tunnels for DatabaseStore reply found");
+					{		
+						// send response through exploratory tunnel
+						auto pool = i2p::tunnel::tunnels.GetExploratoryPool ();
+						auto outbound = pool ? pool->GetNextOutboundTunnel () : nullptr;
+						if (outbound)
+							outbound->SendTunnelDataMsgTo (replyIdent, tunnelID, deliveryStatus);
+						else
+							LogPrint (eLogWarning, "NetDb: No outbound tunnels for DatabaseStore reply found");
+					}		
 				}
 			}
 			offset += 32;
@@ -855,19 +879,22 @@ namespace data
 			{
 				if (storeType == NETDB_STORE_TYPE_LEASESET) // 1
 				{
-					LogPrint (eLogDebug, "NetDb: Store request: LeaseSet for ", ident.ToBase32());
+					if (CheckLogLevel (eLogDebug))
+						LogPrint (eLogDebug, "NetDb: Store request: LeaseSet for ", ident.ToBase32());
 					updated = AddLeaseSet (ident, buf + offset, len - offset);
 				}
 				else // all others are considered as LeaseSet2
 				{
-					LogPrint (eLogDebug, "NetDb: Store request: LeaseSet2 of type ", int(storeType), " for ", ident.ToBase32());
+					if (CheckLogLevel (eLogDebug))
+						LogPrint (eLogDebug, "NetDb: Store request: LeaseSet2 of type ", int(storeType), " for ", ident.ToBase32());
 					updated = AddLeaseSet2 (ident, buf + offset, len - offset, storeType);
 				}
 			}
 		}
 		else // RouterInfo
 		{
-			LogPrint (eLogDebug, "NetDb: Store request: RouterInfo");
+			if (CheckLogLevel (eLogDebug))
+				LogPrint (eLogDebug, "NetDb: Store request: RouterInfo ", ident.ToBase64());
 			size_t size = bufbe16toh (buf + offset);
 			offset += 2;
 			if (size > MAX_RI_BUFFER_SIZE || size > len - offset)
@@ -918,39 +945,9 @@ namespace data
 		auto dest = m_Requests.FindRequest (ident);
 		if (dest)
 		{
-			bool deleteDest = true;
 			if (num > 0)
-			{
-				auto pool = i2p::tunnel::tunnels.GetExploratoryPool ();
-				auto outbound = pool ? pool->GetNextOutboundTunnel () : nullptr;
-				auto inbound = pool ? pool->GetNextInboundTunnel () : nullptr;
-				if (!dest->IsExploratory ())
-				{
-					// reply to our destination. Try other floodfills
-					if (outbound && inbound)
-					{
-						auto count = dest->GetExcludedPeers ().size ();
-						if (count < 7)
-						{
-							auto nextFloodfill = GetClosestFloodfill (dest->GetDestination (), dest->GetExcludedPeers ());
-							if (nextFloodfill)
-							{
-								// request destination
-								LogPrint (eLogDebug, "NetDb: Try ", key, " at ", count, " floodfill ", nextFloodfill->GetIdentHash ().ToBase64 ());
-								outbound->SendTunnelDataMsgTo (nextFloodfill->GetIdentHash (), 0,
-									dest->CreateRequestMessage (nextFloodfill, inbound));
-								deleteDest = false;
-							}
-						}
-						else
-							LogPrint (eLogWarning, "NetDb: ", key, " was not found on ", count, " floodfills");
-					}
-				}
-
-				if (deleteDest)
-					// no more requests for the destinationation. delete it
-					m_Requests.RequestComplete (ident, nullptr);
-			}
+				// try to send next requests
+				m_Requests.SendNextRequest (dest);
 			else
 				// no more requests for destination possible. delete it
 				m_Requests.RequestComplete (ident, nullptr);
@@ -974,8 +971,10 @@ namespace data
 				LogPrint (eLogDebug, "NetDb: Found new/outdated router. Requesting RouterInfo...");
 				if(m_FloodfillBootstrap)
 					RequestDestinationFrom(router, m_FloodfillBootstrap->GetIdentHash(), true);
-				else
+				else if (!IsRouterBanned (router))
 					RequestDestination (router);
+				else
+					LogPrint (eLogDebug, "NetDb: Router ", peerHash, " is banned. Skipped");
 			}
 			else
 				LogPrint (eLogDebug, "NetDb: [:|||:]");
@@ -1019,6 +1018,11 @@ namespace data
 		std::shared_ptr<I2NPMessage> replyMsg;
 		if (lookupType == DATABASE_LOOKUP_TYPE_EXPLORATORY_LOOKUP)
 		{
+			if (!context.IsFloodfill ())
+			{
+				LogPrint (eLogWarning, "NetDb: Exploratory lookup to non-floodfill dropped");
+				return;
+			}	
 			LogPrint (eLogInfo, "NetDb: Exploratory close to ", key, " ", numExcluded, " excluded");
 			std::set<IdentHash> excludedRouters;
 			const uint8_t * excluded_ident = excluded;
@@ -1044,6 +1048,7 @@ namespace data
 			if (lookupType == DATABASE_LOOKUP_TYPE_ROUTERINFO_LOOKUP ||
 				lookupType == DATABASE_LOOKUP_TYPE_NORMAL_LOOKUP)
 			{
+				// try to find router
 				auto router = FindRouter (ident);
 				if (router && !router->IsUnreachable ())
 				{
@@ -1056,17 +1061,26 @@ namespace data
 			if (!replyMsg && (lookupType == DATABASE_LOOKUP_TYPE_LEASESET_LOOKUP ||
 				lookupType == DATABASE_LOOKUP_TYPE_NORMAL_LOOKUP))
 			{
-				auto leaseSet = FindLeaseSet (ident);
-				if (!leaseSet)
+				// try to find leaseset
+				if (context.IsFloodfill ())
+				{	
+					auto leaseSet = FindLeaseSet (ident);
+					if (!leaseSet)
+					{
+						// no leaseset found
+						LogPrint(eLogDebug, "NetDb: Requested LeaseSet not found for ", ident.ToBase32());
+					}
+					else if (!leaseSet->IsExpired ()) // we don't send back expired leasesets
+					{
+						LogPrint (eLogDebug, "NetDb: Requested LeaseSet ", key, " found");
+						replyMsg = CreateDatabaseStoreMsg (ident, leaseSet);
+					}
+				}	
+				else if (lookupType == DATABASE_LOOKUP_TYPE_LEASESET_LOOKUP)
 				{
-					// no lease set found
-					LogPrint(eLogDebug, "NetDb: Requested LeaseSet not found for ", ident.ToBase32());
-				}
-				else if (!leaseSet->IsExpired ()) // we don't send back our LeaseSets
-				{
-					LogPrint (eLogDebug, "NetDb: Requested LeaseSet ", key, " found");
-					replyMsg = CreateDatabaseStoreMsg (ident, leaseSet);
-				}
+					LogPrint (eLogWarning, "NetDb: Explicit LeaseSet lookup to non-floodfill dropped");
+					return;
+				}	
 			}
 
 			if (!replyMsg)
@@ -1114,12 +1128,24 @@ namespace data
 					else
 						LogPrint(eLogWarning, "NetDb: Encrypted reply requested but no tags provided");
 				}
-				auto exploratoryPool = i2p::tunnel::tunnels.GetExploratoryPool ();
-				auto outbound = exploratoryPool ? exploratoryPool->GetNextOutboundTunnel () : nullptr;
-				if (outbound)
-					outbound->SendTunnelDataMsgTo (replyIdent, replyTunnelID, replyMsg);
-				else
+				bool direct = true;
+				if (!i2p::transport::transports.IsConnected (replyIdent))
+				{
+					auto r = FindRouter (replyIdent);
+					if (r && !r->IsReachableFrom (i2p::context.GetRouterInfo ()))
+						direct = false;
+				}	
+				if (direct)
 					transports.SendMessage (replyIdent, i2p::CreateTunnelGatewayMsg (replyTunnelID, replyMsg));
+				else
+				{	
+					auto exploratoryPool = i2p::tunnel::tunnels.GetExploratoryPool ();
+					auto outbound = exploratoryPool ? exploratoryPool->GetNextOutboundTunnel () : nullptr;
+					if (outbound)
+						outbound->SendTunnelDataMsgTo (replyIdent, replyTunnelID, replyMsg);
+					else
+						LogPrint (eLogWarning, "NetDb: Can't send lookup reply to ", replyIdent.ToBase64 (), ". Non reachable and no outbound tunnels");
+				}	
 			}
 			else
 				transports.SendMessage (replyIdent, replyMsg);
@@ -1140,7 +1166,7 @@ namespace data
 		for (int i = 0; i < numDestinations; i++)
 		{
 			RAND_bytes (randomHash, 32);
-			auto dest = m_Requests.CreateRequest (randomHash, true); // exploratory
+			auto dest = m_Requests.CreateRequest (randomHash, true, !throughTunnels); // exploratory
 			if (!dest)
 			{
 				LogPrint (eLogWarning, "NetDb: Exploratory destination is requested already");
@@ -1260,10 +1286,10 @@ namespace data
 	std::shared_ptr<const RouterInfo> NetDb::GetRandomRouter (Filter filter) const
 	{
 		if (m_RouterInfos.empty())
-			return 0;
+			return nullptr;
 		uint16_t inds[3];
 		RAND_bytes ((uint8_t *)inds, sizeof (inds));
-		std::unique_lock<std::mutex> l(m_RouterInfosMutex);
+		std::lock_guard<std::mutex> l(m_RouterInfosMutex);
 		auto count = m_RouterInfos.size ();
 		if(count == 0) return nullptr;
 		inds[0] %= count;
@@ -1325,7 +1351,7 @@ namespace data
 		const std::set<IdentHash>& excluded) const
 	{
 		IdentHash destKey = CreateRoutingKey (destination);
-		std::unique_lock<std::mutex> l(m_FloodfillsMutex);
+		std::lock_guard<std::mutex> l(m_FloodfillsMutex);
 		return m_Floodfills.FindClosest (destKey, [&excluded](const std::shared_ptr<RouterInfo>& r)->bool
 			{
 				return r && !r->IsUnreachable () && !r->GetProfile ()->IsUnreachable () &&
@@ -1340,7 +1366,7 @@ namespace data
 		IdentHash destKey = CreateRoutingKey (destination);
 		std::vector<std::shared_ptr<RouterInfo> > v;
 		{
-			std::unique_lock<std::mutex> l(m_FloodfillsMutex);
+			std::lock_guard<std::mutex> l(m_FloodfillsMutex);
 			v = m_Floodfills.FindClosest (destKey, num, [&excluded](const std::shared_ptr<RouterInfo>& r)->bool
 				{
 					return r && !r->IsUnreachable () && !r->GetProfile ()->IsUnreachable () &&
@@ -1395,7 +1421,7 @@ namespace data
 	{
 		auto ts = i2p::util::GetSecondsSinceEpoch ();
 		{
-			std::unique_lock<std::mutex> l(m_RouterInfosMutex);
+			std::lock_guard<std::mutex> l(m_RouterInfosMutex);
 			for (auto& it: m_RouterInfos)
 				it.second->UpdateIntroducers (ts);
 		}
